/***********************************************************************
|  Enhanced C library
|  Copyright (C) Karl Dahlke 1994
|  This software may be freely used by anyone for any personal,
|  non-profit purpose, and may be freely distributed to others.
|*********************************************************************/

/* index.c: manage linear indexes for serial.c */

#include <errno.h>

#include "eclib.h"

#define CHFD_MAX 15

extern char indexdir[];

/* static variables to communicate with index updates */
static long i_hole; /* where to write the next entry */
static long i_match; /* was this entry found */
static long i_last; /* last entry with this key */
static long i_almost_last; /* next to last, incase we need to delete the last one */
long i_count; /* for statistics */
static long ks[2];

/* Hold files open, cache the open file descriptors.
 * Ties into the eopen() routine. */
static int n_chfd = 0;
static short chfd[CHFD_MAX];
static char chfd_idx[CHFD_MAX], chfd_num[CHFD_MAX];

static index_lookup(idxnum, fd, key, rowid)
long key, rowid;
{
int n;
long offset;
char last;

i_hole = i_match = i_last = i_almost_last = -1;
i_count = 0;

offset = (key * idx_desc[idxnum].numer / idx_desc[idxnum].denom) * 8;
/* this offset may be way beyond the end of the file. */
file_extend(fd, offset, 0L);
elseek(fd, offset, 0);

while(n = eread(fd, ks, 8)) {
last = 0;
++i_count;
offset += 8;
if(!ks[0]) break; /* hole */
if(ks[0] < 0) last = 1, ks[0] = -ks[0];
if(ks[0] == AVAIL_CONSTANT) {
if(i_hole < 0) i_hole = offset-8;
continue;
}
if(ks[0] != key) continue;
if(ks[1] == rowid) i_match = offset-8;
if(last) i_last = offset-8;
else i_almost_last = offset-8;
} /* until eof */

if(i_hole < 0) {
i_hole = offset;
if(n) i_hole -= 8;
}
} /* index_lookup */

/* add an index value */
static addind(idxnum, fd, key, rowid)
long key, rowid;
{

index_lookup(idxnum, fd, key, rowid);

/* if no longer last, back up and reverse the sign */
if(i_last >= 0 && i_hole > i_last) {
elseek(fd, i_last, 0);
eread(fd, ks, 8);
ks[0] = -ks[0];
elseek(fd, i_last, 0);
ewrite(fd, ks, 8);
}

/* now insert into the index */
ks[0] = i_last > i_hole ? key : -key;
ks[1] = rowid;
elseek(fd, i_hole, 0);
ewrite(fd, ks, 8);
} /* addind */

/* remove an index entry */
static rmind(idxnum, fd, key, rowid)
long key, rowid;
{
index_lookup(idxnum, fd, key, rowid);
if(i_match < 0 || i_last < 0)
eprint("@rmind index match or last not set");

/* match may coincide with last, almost_last, or something earlier.
 * If there is an almost_last, not equal to match,
 * then we must make this look like the last element with this key. */
if(i_almost_last >= 0 && i_almost_last != i_match) {
/* next-to-last is now last */
elseek(fd, i_almost_last, 0);
eread(fd, ks, 8);
ks[0] = -ks[0];
elseek(fd, i_almost_last, 0);
ewrite(fd, ks, 8);
}

if(i_match != i_last) {
/* move last to match */
elseek(fd, i_last, 0);
eread(fd, ks, 8);
if(i_match != i_almost_last) ks[0] = -ks[0];
elseek(fd, i_match, 0);
ewrite(fd, ks, 8);
}

/* put a temp hole in the last slot */
ks[0] = AVAIL_CONSTANT;
elseek(fd, i_last, 0);
ewrite(fd, ks, 8);
} /* rmind */

/* fix up indexes after insert/update/delete */
index_fixup(oldbuf, newbuf, oldid, newid, relnum)
char *oldbuf, *newbuf;
long oldid, newid;
short relnum;
{
short fd;
struct IDX_DESC *s;
long save_match;
char *t;
short idxnum;
long oldkey, newkey;

for(idxnum=0, s=idx_desc; s->cls; ++s, ++idxnum) {
if(s->relnum != relnum) continue;
oldkey = newkey = 0;

if(oldid) {
t = oldbuf + s->offset;
if(s->cls == 2) oldkey = *((short *)t);
if(s->cls == 4) oldkey = *((long *)t);
}

if(newid) {
t = newbuf + s->offset;
if(s->cls == 2) newkey = *((short *)t);
if(s->cls == 4) newkey = *((long *)t);
}

if(oldkey == newkey) continue;

fd = ifd_setup(idxnum);
save_match = -1L;

if(oldkey) {
rmind(idxnum, fd, oldkey, oldid);
save_match = i_match;
}
if(newkey)
addind(idxnum, fd, newkey, newid);

/* In case we are looping on this index, and making changes to this
 * indexed column, or deleeting rows, reset the index */
if(s->holdopen && save_match >= 0)
elseek(fd, save_match, 0);
} /* end loop over indexes */
} /* index_fixup */

idx_by_name(relnum, idxname)
char *idxname;
{
struct IDX_DESC *idxblock;

for(idxblock = idx_desc; idxblock->cls; ++idxblock) {
if(idxblock->relnum != relnum) continue;
if(strcmp(idxblock->member, idxname)) continue;
return idxblock - idx_desc;
}
eprint("@cannot find index %d.%s", relnum, idxname);
} /* idx_by_name */

cache_fd(fd, isidx, offset)
short fd;
char isidx, offset;
{
int i;

for(i=0; i<n_chfd; ++i)
if(chfd[i] == fd) break;

if(i == CHFD_MAX) {
uncache_fd();
--i;
}

if(i == n_chfd)   ++n_chfd;

while(--i >= 0) {
chfd[i+1] = chfd[i];
chfd_idx[i+1] = chfd_idx[i];
chfd_num[i+1] = chfd_num[i];
}

chfd[0] = fd;
chfd_idx[0] = isidx;
chfd_num[0] = offset;
} /* cache_fd */

uncache_fd()
{
int i, n;
short fd;

for(i=n_chfd-1; i>=0; --i) {
fd = chfd[i];
n = chfd_num[i];

if(chfd_idx[i]) {
if(idx_desc[n].fd != fd)
eprint("@cache fd mismatch in %d.%d", chfd_idx[i], n);
if(idx_desc[n].holdopen) continue;
idx_desc[n].fd = -1;
} else {
if(struct_desc[n].fd != fd)
eprint("@cache fd mismatch in %d.%d", chfd_idx[i], n);
if(struct_desc[n].holdopen) continue;
struct_desc[n].fd = -1;
}
close(fd);
break;
}

if(i < 0)
eprint("@cannot uncache file descriptor, %d files held open", n_chfd);

for(++i; i<n_chfd; ++i) {
chfd[i-1] = chfd[i];
chfd_idx[i-1] = chfd_idx[i];
chfd_num[i-1] = chfd_num[i];
}
--n_chfd;
} /* uncache_fd */

uncache_fd_all()
{
int i, n;

for(i=0; i<n_chfd; ++i) {
n = chfd_num[i];
close(chfd[i]);
if(chfd_idx[i]) {
idx_desc[n].fd = -1;
idx_desc[n].holdopen = 0;
} else {
struct_desc[n].fd = -1;
struct_desc[n].holdopen = 0;
}
}
n_chfd = 0;
} /* uncache_fd_all */

/* drop or update linked tuples, like the cascaded update/delete.
 * This, and similar code, works because we are deleting/selecting/updating the current row.
 * You are allowed to loop through a relation, sequential or indexed,
 * and make updates/deletes, provided you always modify the current row. */
casc_upd_del(key, newkey, relnum, idxcol)
long key, newkey;
int relnum;
char *idxcol; /* indexed column */
{
struct rel_cursor c;
struct IDX_DESC *idxblock;
char *t;
char buf[TUPLEN];

if(key == newkey) return;

idxblock = idx_desc + idx_by_name(relnum, idxcol);
t = buf + idxblock->offset;

start_scan(&c, NULL, relnum, buf, key, idxcol);
while(next_scan(&c)) {
if(!newkey) {
ZC_delete_row_file(relnum, c.rowid);
continue;
}

/* set value to newkey and update */
if(idxblock->cls == 2)
*((short *)t) = newkey;
else
*((long *)t) = newkey;
ZC_insupd_row_file(relnum, c.rowid, buf);
}
end_scan(&c);
} /* casc_upd_del */

reindex(idxnum)
{
int ifd, ofd; /* file descriptors */
int relnum = idx_desc[idxnum].relnum;
long kbuf[2];
char *newindname, *tempidxname();

uncache_fd_all();
lock_set(relnum);
indexdir[strlen(indexdir)-1] = idxnum + 'a';
newindname = tempidxname();
erename(indexdir, newindname);
ifd = eopen(newindname, O_RDONLY|O_BINARY);
ofd = eopen(indexdir, O_RDWR|O_BINARY|O_TRUNC|O_CREAT, 0666);

while(eread(ifd, kbuf, 8)) {
if(kbuf[0] == 0) continue;
if(kbuf[0] == AVAIL_CONSTANT) continue;
if(kbuf[0] < 0) kbuf[0] = - kbuf[0];
addind(idxnum, ofd, kbuf[0], kbuf[1]);
} /* end reading old keys */

close(ifd);
close(ofd);
idxval_write();
unlink(newindname);
lock_clear(relnum);
} /* reindex */

eval_index(update)
{
struct IDX_DESC *ib;
int ifd; /* index file descriptor */
int idxnum;
long holecount, totalcount;
int perc; /* percentage of holes */
long numer, denom;

for(idxnum=0, ib=idx_desc; ib->cls; ++ib, ++idxnum) {
ifd = ifd_setup(idxnum);
totalcount = filesize(ifd) / 8;
if(!totalcount) {
printf("%s.%s is empty\n",
struct_desc[ib->relnum].name, ib->member);
continue;
}
holecount = 0;
elseek(ifd, 0L, 0);

while(eread(ifd, ks, 8)) {
if(ks[0] == 0 || ks[0] == AVAIL_CONSTANT)
++holecount;
}

/* compute percentage of holes */
perc = holecount*100/totalcount;
printf("%s.%s: %2d%%\n",
struct_desc[ib->relnum].name, ib->member, perc);

/* could be all holes, if you add a row and then delete it */
if(perc == 100) continue;
if(ib->bounded) {
if(perc >= 70 && perc < 90) continue;
} else {
if(perc >= 5 && perc < 30) continue;
}

/* ideal percentages are 20 % and 80%, adjust ratio */
numer = ib->numer;
denom = ib->denom;
numer *= (100-perc);
if(!perc) numer *= 3;
denom *= 20;
if(!ib->bounded)
denom *= 4;

while(!((numer%2) | (denom%2)))
numer /= 2, denom /= 2;
while(!((numer%3) | (denom%3)))
numer /= 3, denom /= 3;
while(!((numer%5) | (denom%5)))
numer /= 5, denom /= 5;
while(numer > 12 && denom > 30000 ||
denom > 12 && numer > 500)
numer /= 2, denom /= 2;
printf("Recalibrating linear index from %d/%d to %ld/%ld\n",
ib->numer, ib->denom, numer, denom);
if(!update) continue;
if(numer > 500 || denom > 30000) {
printf("Ratio contains a component integer too large, cannot adjust index.\n");
continue;
}

if(!update) continue;
ib->numer = numer;
ib->denom = denom;
reindex(idxnum);
} /* loop over indexes */
} /* eval_index */

idxval_read()
{
int fd;
char *s, save_ilr;
struct IDX_DESC *ib;

s = indexdir + strlen(indexdir) - 7;
save_ilr = s[6];
strcpy(s, "ratios");
fd = open(indexdir, O_RDONLY|O_BINARY);
if(fd < 0) {
/* if not present, go with the compiled values. */
/* Thus, disaster can strike if somebody removes this file after reindexing. */
if(errno != ENOENT)
eprint("@cannot open %s", indexdir);
} else {
for(ib=idx_desc; ib->cls; ++ib) {
short ratio[2];
if(read(fd, ratio, sizeof(ratio)) != sizeof(ratio))
eprint("@cannot read %s", indexdir);
ib->numer = ratio[0];
ib->denom = ratio[1];
}
close(fd);
}
strcpy(s, "index_x");
s[6] = save_ilr;
} /* idxval_read */

idxval_write()
{
int fd;
char *s, save_ilr;
struct IDX_DESC *ib;

s = indexdir + strlen(indexdir) - 7;
save_ilr = s[6];
strcpy(s, "ratios");
fd = eopen(indexdir, O_WRONLY|O_BINARY|O_CREAT|O_TRUNC, 0666);
for(ib=idx_desc; ib->cls; ++ib) {
short ratio[2];
ratio[0] = ib->numer;
ratio[1] = ib->denom;
ewrite(fd, ratio, sizeof(ratio));
}
close(fd);
strcpy(s, "index_*");
s[6] = save_ilr;
} /* idxval_write */

